"""
## 关键词列表
### 流程关键词(按执行顺序)
- from|name
- 根据key/value依次执行map
- filter
- sort,reverse
- only|(except,rename)
- item

### 变量关键词
- it
- query
- path
- header
- body
- lib
- operation
  - `operation('get-file-accessToken').example('ok')`
  - `operation('get-file-accessToken').example()`
"""
from typing import List, Dict, Any
import json
import re
from wax.load_func import lib


KEYWORDS = [
    '__helper__',
    '__from__',
    '__name__',
    '__filter__',
    '__sort__',
    '__reverse__',
    '__only__',
    '__except__',
    '__rename__',
    '__item__',
]


class PqlRuntimeError(Exception):
    def __init__(self, path, reason):
        self.path = path
        self.reason = reason

    def __str__(self):
        return f'{self.path}: {self.reason}'


def is_var_name(name: str) -> bool:
    return bool(re.match(r'^\w+$', name))


def match_name_pair(text: str) -> Any:
    match_ret = re.match(r'^(\w+:)?(\w+)$', text)
    if not match_ret:
        return None, None
    outer_part, inner_part = match_ret.groups()
    if not outer_part:
        return outer_part, inner_part
    else:
        return outer_part[:-1], inner_part


def query_entity(entity_name) -> List:
    try:
        text = open(f'entity/{entity_name}.json').read()
    except:
        raise PqlRuntimeError('', f'数据未创建({entity_name})')
    try:
        data = json.loads(text)
    except:
        raise PqlRuntimeError('', f'数据无法解析JSON({entity_name})')
    if not isinstance(data, list) or not all(isinstance(item, dict) for item in data):
        raise PqlRuntimeError('', f'数据不合法，只支持list[dict]类型({entity_name})')
    return data


def helper_schema(helper_name, dict_schema: Dict) -> Dict:
    try:
        text = open(f'helper/{helper_name}.json').read()
    except:
        raise PqlRuntimeError('__helper__', f'helper未创建({helper_name})')
    try:
        helper_schema = json.loads(text)
    except:
        raise PqlRuntimeError('__helper__', f'helper无法解析JSON({helper_name})')
    if not isinstance(helper_schema, dict):
        raise PqlRuntimeError('__helper__', f'helper不合法，只支持dict类型({helper_name})')
    for key, value in dict_schema.items():
        if key == '__helper__':
            continue
        helper_schema[key] = value
    return helper_schema


def apply_schema(env, dict_schema) -> Any:
    """
    把dict类型的schema实例化为数组类型的rows，或__return__指定的值
    """
    if not isinstance(dict_schema, dict):
        raise PqlRuntimeError('', 'schema必须是dict类型')
    # 处理helper类型的schema
    if '__helper__' in dict_schema:
        filled_schema = helper_schema(helper_name=dict_schema['__helper__'], dict_schema=dict_schema)
        return apply_schema(env, dict_schema=filled_schema)
    filterd_rows = []
    renamed_rows = []
    outer_name, cur_rows = None, [{}]
    # 第一阶段：name|from
    if '__name__' in dict_schema and '__from__' in dict_schema:
        raise PqlRuntimeError('__name__', '__name__和__from__不能同时定义')
    if '__name__' in dict_schema:
        if not is_var_name(dict_schema['__name__']):
            raise PqlRuntimeError('__name__', '语法错误')
        outer_name = dict_schema['__name__']
    elif '__from__' in dict_schema:
        outer_name, inner_name = match_name_pair(dict_schema['__from__'])
        if not inner_name:
            raise PqlRuntimeError('__from__', '语法错误')
        try:
            cur_rows = query_entity(inner_name)
        except PqlRuntimeError as e:
            raise PqlRuntimeError('__from__', e.reason)
    for cur_row in cur_rows:
        env['it'] = cur_row
        if outer_name:
            env[outer_name] = cur_row
        # 第二阶段：根据key/value依次执行map
        for key, func_or_schemas in dict_schema.items():
            if key in KEYWORDS:
                continue
            value = None
            if not isinstance(func_or_schemas, list):
                func_or_schemas = [func_or_schemas]
            for func_or_schema in func_or_schemas:
                if isinstance(func_or_schema, str):
                    value = apply_lambda(env, func=func_or_schema, key=key)
                elif isinstance(func_or_schema, dict):
                    try:
                        value = apply_schema(dict(env), dict_schema=func_or_schema)
                    except PqlRuntimeError as e:
                        raise PqlRuntimeError(f'{key}/{e.path}', e.reason)
                else:
                    raise PqlRuntimeError(key, '语法错误，只支持dict或str类型')
            try:
                cur_row.update(apply_items(key, rows=value))
            except PqlRuntimeError as e:
                raise PqlRuntimeError(f'{key}', e.reason)
        # 第三阶段：filter
        is_chosen = True
        if '__filter__' in dict_schema:
            func = dict_schema['__filter__']
            is_chosen = apply_lambda(env, func, key='__filter__')
        if is_chosen:
            filterd_rows.append(cur_row)
    # 第四阶段：sort,reverse
    need_reverse, need_sort = False, False
    if '__reverse__' in dict_schema:
        rev_func = dict_schema['__reverse__']
        rev_env = {k: v for (k, v) in env.items() if k != 'it'}  # 排除it
        need_reverse = bool(apply_lambda(rev_env, func=rev_func, key='__reverse__'))
    if '__sort__' in dict_schema:
        sort_func = dict_schema['__sort__']
        need_sort = True
    if need_sort:
        filterd_rows.sort(key=lambda it: apply_lambda(dict(env, it=it), sort_func, key='__sort__'), reverse=need_reverse)
    elif need_reverse:
        filterd_rows.reverse()
    # 第五阶段：only|(except,rename)
    # 本阶段的产出是renamed_rows
    only_names, except_names = [], []
    rename_rules = {}
    if '__only__' in dict_schema:
        if '__except__' in dict_schema:
            raise PqlRuntimeError('__only__', '__only__和__except__不能同时定义')
        if '__rename__' in dict_schema:
            raise PqlRuntimeError('__only__', '__only__和__rename__不能同时定义')
        rename_pairs = dict_schema['__only__']
        if not isinstance(rename_pairs, list) or not all(isinstance(name, str) for name in rename_pairs):
            raise PqlRuntimeError('__only__', '语法错误，只支持list[str]类型')
        for name_pair in rename_pairs:
            outer_name, inner_name = match_name_pair(name_pair)
            if not inner_name:
                raise PqlRuntimeError('__only__', '语法错误')
            only_names.append(inner_name)
            if outer_name:
                rename_rules[inner_name] = outer_name
    elif '__except__' in dict_schema:
        except_names.extend(dict_schema['__except__'])
        if not isinstance(except_names, list) or not all(isinstance(name, str) for name in except_names):
            raise PqlRuntimeError('__except__', '语法错误，只支持list[str]类型')
        if not all(is_var_name(name) for name in except_names):
            raise PqlRuntimeError('__except__', '语法错误')
    if '__rename__' in dict_schema:
        rename_pairs = dict_schema['__rename__']
        if not isinstance(rename_pairs, list) or not all(isinstance(name, str) for name in rename_pairs):
            raise PqlRuntimeError('__rename__', '语法错误，只支持list[str]类型')
        for name_pair in rename_pairs:
            outer_name, inner_name = match_name_pair(name_pair)
            if not inner_name or not outer_name:
                raise PqlRuntimeError('__rename__', '语法错误')
            rename_rules[inner_name] = outer_name
    for cur_row in filterd_rows:
        if isinstance(cur_row, dict):
            # 处理only|except
            if '__only__' in dict_schema:  # 重要，不可省略！
                cur_row = {key: value for (key, value) in cur_row.items() if key in only_names}
            elif '__except__' in dict_schema:
                cur_row = {key: value for (key, value) in cur_row.items() if key not in except_names}
            # 处理rename
            renamed_row = {}
            for key, value in cur_row.items():
                if key in rename_rules:
                    renamed_row[rename_rules[key]] = value
                else:
                    renamed_row[key] = value
            # 排除结果中的None值
            renamed_row = {key: value for (key, value) in renamed_row.items()
                           if key and value is not None}
            renamed_rows.append(renamed_row)
    # 第六阶段：item
    if '__item__' in dict_schema:
        item_indices = dict_schema['__item__']
        if not isinstance(item_indices, list) or not all(
                indice is None or isinstance(indice, int) for indice in item_indices):
            raise PqlRuntimeError('__return__', '语法错误，只支持list[int?]类型')
        if not 1 <= len(item_indices) <= 3:
            raise PqlRuntimeError('__return__', '语法错误，长度应该在1~3范围内')
        if len(item_indices) == 1 and item_indices[0] is None:
            raise PqlRuntimeError('__return__', '语法错误，单个下标不能为null')
        try:
            if len(item_indices) == 1:
                return renamed_rows[item_indices[0]]
            elif len(item_indices) == 2:
                start, end = item_indices
                return renamed_rows[start:end]
            elif len(item_indices) == 3:
                start, end, step = item_indices
                return renamed_rows[start:end:step]
        except:
            # 此处故意不抛异常，理解为None是rows[0]的默认值
            return None
    else:
        return renamed_rows


def apply_items(key: str, rows: Any) -> Dict:
    """
    把(key, rows)变成字典类型的值。
    例如：
    {'name': rows} => {'name': rows}
    {'**name': row} => {**row}
    {'**name': rows} => {**rows[0]}
    {'**name[0]': rows} => {}  # error!
    """
    match_ret = re.match(r'^(\*\*)?(\w+)$', key)
    if not match_ret:
        raise PqlRuntimeError('', 'key语法错误')
    if key.startswith('**'):
        if isinstance(rows, list):
            if not rows:
                return {}
            row = rows[0]
        else:
            row = rows
        if row is None:  # {"__item__": [0]}可能返回None
            return {}
        if not isinstance(row, dict):
            raise PqlRuntimeError('', '运行时错误，只支持dict?|list[dict?]类型')
        key_prefix = match_ret.groups()[1]
        return {
            (key_prefix + sub_key if key_prefix[-1] == '_' else key_prefix + sub_key[0].upper() + sub_key[1:]): value
            for sub_key, value in row.items()
        }
    else:
        return {key: rows}


def apply_lambda(env: Dict, func: str, key: str) -> Any:
    if not isinstance(func, str):
        raise PqlRuntimeError(key, '语法错误，只支持str类型')
    if not func:
        raise PqlRuntimeError(key, '语法错误，不支持空字符串')
    try:
        return eval('lambda %s: (%s)' % (','.join(env), func))(**env)
    except SyntaxError as e:
        raise PqlRuntimeError(key, '语法错误: ' + str(e))
    except Exception as e:
        raise PqlRuntimeError(key, f'运行时错误: ({type(e).__name__}) {e}')



class SortedDict:
    def of(self, dict_data):
        ret = {}
        for key, val in sorted(dict_data.items()):
            ret[key] = val
        return ret

    def __matmul__(self, other):
        return self.of(other)

    def __rmatmul__(self, other):
        return self.of(other)


sorted_dict = SortedDict()



from typing import Dict, List
import itertools
import json
from mako.template import Template  # type: ignore
from wax.lessweb import BadParamError, Context
from wax.lessweb.webapi import http_methods
from wax.service import StateServ
from wax.load_config import config
from wax.load_swagger import SwaggerData, parse_operation
from wax.jsonschema_util import compare_json
from wax.kotlin_util import import_headers, schema_to_kclass, endpoint_to_kcontroller, kclass_index


def split_tag(tag: str) -> List[str]:
    if not tag:
        tag = '其他'
    segs = tag.replace('/', '-').split('-', 2)
    if len(segs) == 1:
        segs.append('API')
    if len(segs) == 2:
        segs.append('API')
    return segs


class Operation:
    method: str
    path: str
    summary: str
    operationId: str
    description: str
    # state
    basic: str
    extra: str
    all_example: List[str]  # List[stateBasic]


# {majorTag: {dirTag: {menuTag: List[Operation]}}}
tag_tree: Dict[str, Dict[str, Dict[str, List[Operation]]]] = {}
op_index: Dict[str, Operation] = {}  # {operationId: Operation}


def first_time_append_op(tag_set, major_tag, op) -> bool:
    tag_set.setdefault(major_tag, set())
    if (op.method, op.path) not in tag_set[major_tag]:
        tag_set[major_tag].add((op.method, op.path))
        return True
    else:
        return False


def load_tag():
    tag_set = {}
    swagger_data = SwaggerData.get()
    for endpoint_path, endpoint in swagger_data['paths'].items():
        for op_method, operation in endpoint.items():
            if op_method.upper() not in http_methods: continue
            for tag in operation.get('tags', []) or ['']:
                major_tag, dir_tag, menu_tag = split_tag(tag)
                tag_tree.setdefault(major_tag, {})
                tag_tree[major_tag].setdefault(dir_tag, {})
                tag_tree[major_tag][dir_tag].setdefault(menu_tag, [])
                op = Operation()
                op.method = op_method.upper()
                op.path = endpoint_path
                try:
                    op.summary = operation['summary']
                    op.operationId = operation['operationId']
                except:
                    print(endpoint_path, op_method.upper())
                    raise
                op.description = operation.get('description', '')
                op.all_example = []
                for status_code, response_val in operation.get('responses', {}).items():
                    if not response_val.get('content'): print(f'WARNING: responses is empty -> {op.method} {op.path}')
                    for content_key, content_val in response_val.get('content', {}).items():
                        for example_name, _ in content_val.get('examples', {}).items():
                            op.all_example.append(f'{status_code}:{content_key}:{example_name}')
                tag_tree[major_tag][dir_tag][menu_tag].append(op)
                #
                tag_tree.setdefault('API', {'API': {'API': []}})
                tag_tree[major_tag].setdefault('API', {'API': []})
                if first_time_append_op(tag_set, 'API', op):
                    tag_tree['API']['API']['API'].append(op)
                if first_time_append_op(tag_set, major_tag, op) and dir_tag != 'API':
                    tag_tree[major_tag]['API']['API'].append(op)
                #
                op_index[op.operationId] = op


load_tag()


def operation_list(tag: str='API'):
    if not tag and tag_tree:
        tag = list(tag_tree.keys())[0]
    major_tag, dir_tag, menu_tag = split_tag(tag)
    return Template(filename='wax-www/tpl/tag_list_page.mako', input_encoding='utf-8', output_encoding='utf-8').render(
        tag_tree=tag_tree,
        major_tag=major_tag,
        dir_tag=dir_tag,
        menu_tag=menu_tag,
        git_url=config['git-url'],
        title=config['title']
    )


def operation_detail(opId: str, show: str=''):
    op = op_index.get(opId)
    swagger_data = SwaggerData.get()
    endpoint = [val for key, val in swagger_data['paths'].items() if key==op.path][0]
    operation = endpoint[op.method.lower()]
    data = {'op': operation, 'path': op.path, 'method': op.method, 'mock_prefix': config['mockapi-prefix']}
    data.update(parse_operation(swagger_data, endpoint, op.method))
    if show == 'json':
        return data
    return Template(filename='wax-www/tpl/op_detail_page.mako', input_encoding='utf-8', output_encoding='utf-8').render(
        tag_tree=tag_tree,
        git_url=config['git-url'],
        **data
    )


def operation_example(opId: str, state: StateServ):
    op = op_index.get(opId)
    if not op:
        raise BadParamError(message='opId不存在', param='opId')
    state.operation_id = opId
    op.basic = state.basic.get()
    op.extra = state.extra.get()
    return op


def operation_edit_state(state: StateServ, opId: str, basic:str='', extra:str=''):
    if opId not in op_index:
        raise BadParamError(message='opId不存在', param='opId')
    state.operation_id = opId
    if basic is not None:
        state.basic.set(basic, ex=86400 * 90)
    if extra is not None:
        state.extra.set(extra, ex=86400 * 90)
    return {}


def compare_swagger(ctx: Context, actual: dict, ignore: str='') -> List[str]:
    ret = []
    expect = SwaggerData.get()
    actual_paths = actual.get('paths', {})
    ignores = ignore.split(',') if ignore else []
    for path in actual_paths.keys() | expect['paths'].keys():
        if path in ignores:
            continue
        if path not in actual_paths:
            ret.append(f'actual未包含path: {path}')
            continue
        if path not in expect['paths']:
            ret.append(f'expect未包含path: {path}')
            continue
        for method in actual_paths[path].keys() | expect['paths'][path].keys():
            if method.upper() not in http_methods:
                continue
            if method not in actual_paths[path]:
                ret.append(f'actual未包含接口：{method.upper()} {path}')
                continue
            if method not in expect['paths'][path]:
                ret.append(f'expect未包含接口：{method.upper()} {path}')
                continue
            actual_endpoint = actual_paths[path]
            expect_endpoint = expect['paths'][path]
            actual_op = parse_operation(actual, actual_endpoint, method.upper())
            expect_op = parse_operation(expect, expect_endpoint, method.upper())
            ret.extend(compare_json(f'{path}:{method}:parameters', actual_op['params'], expect_op['params'], actual, expect))
            ret.extend(compare_json(
                f'{path}:{method}:requestBody:content',
                actual_endpoint[method].get('requestBody', {}).get('content'),
                expect_endpoint[method].get('requestBody', {}).get('content'),
                actual, expect))
            try:
                actual_endpoint[method]['responses']['200']['content']['application/json'] = actual_endpoint[
                    method]['responses']['200']['content'].pop('*/*')
            except:
                pass
            ret.extend(compare_json(
                f'{path}:{method}:responses',
                actual_endpoint[method].get('responses'),
                expect_endpoint[method].get('responses'),
                actual, expect))
    return ret


def make_kotlin_code(ctx: Context) -> str:
    swagger_data = SwaggerData.get()
    ret = [import_headers()]
    for name, schema in swagger_data['components']['schemas'].items():
        schema_to_kclass(schema, name, swagger_data)
    for path, endpoint in swagger_data['paths'].items():
        ret.append(endpoint_to_kcontroller(path, endpoint, swagger_data))
    for kclass in kclass_index.values():
        ret.append(kclass.to_class_impl())
        if kclass.generics:
            ret.append(kclass.to_inherit())
    ctx.response.send_content_type('txt', encoding='utf-8')
    return '\n'.join(ret)


def make_solution_list(ctx: Context) -> str:
    swagger_data = SwaggerData.get()
    task_list = []  # (path, method, summary, label)
    methods = ['POST', 'PUT', 'GET', 'DELETE']
    for path, endpoint in swagger_data['paths'].items():
        for method, operation in endpoint.items():
            if method.upper() not in http_methods: continue
            label = '-'.join(path.strip('/').split('/')[:2]) or 'homepath'
            task_list.append((method.upper(), path, operation.get('summary') or operation.get('operationId'), '~' + label))
    task_list.sort(key=lambda x: (x[1], methods.index(x[0])))
    lines = []
    for task in task_list:
        lines.append(' * ' + ' '.join(task) + '\n')
    ctx.response.send_content_type('txt', encoding='utf-8')
    return ''.join(lines)


def make_openapi_json() -> Dict:
    swagger_data = SwaggerData.get()
    cleared_swagger = {
        key: (
            val if key != 'paths' else (
                lambda items: {
                    path: (
                        lambda items: {
                            method: (
                                (
                                    lambda items: {
                                        mkey: (val if mkey != 'tags' else [])
                                        for mkey, val in items
                                    }
                                )(val.items())
                                if isinstance(val, dict) else val
                            ) for method, val in items}
                    )(val.items())
                for path, val in items}
            )(val.items())
        ) for key, val in swagger_data.items()}
    return cleared_swagger



from typing import List, Optional, Dict, Union
import json
import re
import hashlib
import base64
from wax.lessweb.webapi import http_methods, BadParamError
from wax.jsonschema_util import jsonschema_from_ref
from wax.load_swagger import parse_operation


def indented(text: str, n=4):
    ret = '\n'.join(' ' * n + line for line in text.splitlines())
    if text.endswith('\n'):
        ret += '\n'
    return ret


def safe_str(text: str) -> str:
    return json.dumps(text, ensure_ascii=False)


def letter_sign(text: str) -> str:
    """生成base32编码的MD5签名"""
    return base64.b32encode(hashlib.md5(text.encode()).digest()).decode()


def properties_sign(properties: Dict, required: List) -> str:
    core_list = [(key, val.get('format') or val.get('type'), val.get('description') or '')
                 for key, val in properties.items()]
    core_list.sort()
    core_list.extend(sorted(required))
    return letter_sign(json.dumps(core_list, ensure_ascii=False))


def capitalize(text: str) -> str:
    """首字母大写，其他字母不变"""
    return text[0].upper() + text[1:] if text else ''


def make_funcname(opId: str) -> str:
    opId = re.sub("[_/{}\"']", '-', opId)
    ret = ''.join(capitalize(seg) for seg in opId.split('-'))
    return ret[0].lower() + ret[1:]


def make_typename(prefix: str, schema):
    properties = schema.get('properties', {})
    required = schema.get('required', [])
    if not properties:
        prefix = 'Success'
    elif 'pageNum' in properties and 'total' in properties and 'list' in properties:
        prefix = 'CommonPage'
    ret = capitalize(prefix)
    sign = properties_sign(properties, required)
    ret += 'T' + sign[:4].lower()
    return ret + 'Vo'


def jsontype_to_ktype(jsontype: str, *, format: str=None) -> str:
    if jsontype == 'integer':
        if format == 'int64':
            return 'Long'
        else:
            return 'Int'
    elif jsontype == 'string':
        if format == 'date-time':
            return 'LocalDateTime'
        elif format == 'date':
            return 'LocalDate'
        elif format == 'time':
            return 'LocalTime'
        else:
            return 'String'
    elif jsontype == 'number':
        return 'Double'
    elif jsontype == 'boolean':
        return 'Boolean'
    elif jsontype == 'array':
        return 'List'
    elif jsontype == 'object':
        return 'JSONObject'
    else:
        raise NotImplementedError((jsontype, format))


class Kprop:
    name: str
    ktype: str  # 例如：Int, List
    generic: str  # 范型类型名
    notNull: bool
    description: str
    kclass: Optional['Kclass'] = None  # 仅object类型设置kclass
    intMin: Optional[int] = None
    intMax: Optional[int] = None

    def to_ktype(self, type_var='') -> str:
        if self.ktype == 'JSONObject' and self.generic:
            return self.generic
        elif not self.generic:
            return self.ktype
        elif type_var:
            return f'{self.ktype}<{type_var}>'
        else:
            return f'{self.ktype}<{self.generic}>'

    def to_kcode(self, type_var='') -> str:
        real_ktype = self.to_ktype(type_var)
        return ('@NotNull\n' if self.notNull else '') + \
               f"""@Schema(description = {safe_str(self.description)})
var {self.name}: {real_ktype}? = null\n"""


class Kclass:
    typeName: str
    properties: List[Kprop]
    generics: List[str]  # 包含的范型

    def to_inherit(self) -> str:
        assert self.generics
        ret = f'class {self.typeName}{"".join(self.generics)}: {self.typeName}<'
        ret += ', '.join(self.generics)
        ret += '>()\n'
        return ret

    def to_class_impl(self) -> str:
        ret = f'open class {self.typeName}'
        if self.generics:
            ret += '<' + ', '.join(chr(ord('T') + i) for i, _ in enumerate(self.generics)) + '>'
        ret += ' {\n'
        for prop in self.properties:
            ret += indented(prop.to_kcode(type_var='T')) + '\n'
        ret += '}\n'
        return ret


kclass_index: Dict[str, Kclass] = {}  # {sign: Kclass}


def schema_to_kprop(propname: str, schema: Dict, swagger_data) -> Kprop:
    description = schema.get('description', '')
    notNull = False
    kclass = None
    if '$ref' in schema:
        ref_schema = jsonschema_from_ref(schema['$ref'], swagger_data)
        typename = schema['$ref'].rsplit('/', 1)[-1]
        kclass = schema_to_kclass(ref_schema, typename, swagger_data)
        ktype = jsontype_to_ktype('object')
        generic = f'{kclass.typeName}<{", ".join(kclass.generics)}>' if kclass.generics else kclass.typeName
    else:
        types = schema.get('type') or ['string']
        if not isinstance(types, list):
            types = [types]
        if 'array' in types:
            items = schema.get('items', {})
            kprop = schema_to_kprop(propname + 'Item', items, swagger_data)
            ktype = jsontype_to_ktype('array')
            if kprop.generic:
                generic = kprop.generic
            else:
                ktype += '<' + kprop.ktype + '>'
                generic = ''
        elif 'object' in types:
            kclass = schema_to_kclass(schema, make_typename(propname, schema), swagger_data)
            ktype = jsontype_to_ktype('object')
            generic = f'{kclass.typeName}<{", ".join(kclass.generics)}>' if kclass.generics else kclass.typeName
        else:
            ktype = jsontype_to_ktype(schema.get('type'), format=schema.get('format'))
            generic = ''
    kprop = Kprop()
    kprop.name = propname
    kprop.ktype = ktype
    kprop.generic = generic
    kprop.notNull = notNull
    kprop.description = description
    kprop.kclass = kclass
    return kprop


def schema_to_kclass(schema: Dict, typeName, swagger_data) -> Kclass:
    """
    object_type_schema -> properties & required -> kclass
    """
    assert typeName
    properties = schema.get('properties', {})
    required = schema.get('required', [])
    sign = properties_sign(properties, required)
    if sign in kclass_index:
        if typeName in swagger_data['components']['schemas']:
            kclass_index[sign].typeName = typeName
        return kclass_index[sign]
    kclass = Kclass()
    kclass.typeName = typeName
    kclass.properties = []
    kclass.generics = []
    for key, val in properties.items():
        kprop = schema_to_kprop(key, val, swagger_data)
        if key in required:
            kprop.notNull = True
        kclass.properties.append(kprop)
        if kprop.ktype == 'List' and kprop.generic:
            kclass.generics.append(kprop.generic)
    kclass_index[sign] = kclass
    return kclass


def schema_to_votype(name: str, schema: Dict, swagger_data, *, inherit=True) -> str:
    kprop = schema_to_kprop(name, schema, swagger_data)
    if kprop.kclass is not None and kprop.kclass.generics:
        ret = f'{kprop.kclass.typeName}<{", ".join(kprop.kclass.generics)}>'
        if inherit:
            return ''.join(re.sub('[,<>]', '', ret).split())
        else:
            return ret
    else:
        return kprop.to_ktype()


def endpoint_to_kcontroller(path, endpoint, swagger_data) -> str:
    controller_name = capitalize(make_funcname(path))
    ret = f"""@RestController
@Validated
class {controller_name}Controller""" + ' {\n'
    for method, operation in endpoint.items():
        if method.upper() not in http_methods: continue
        parsed_op = parse_operation(swagger_data, endpoint, method)
        op_text = f'@{method.capitalize()}Mapping("{path}")\n'
        op_summary = operation.get('summary', '')
        op_description = operation.get('description', '')
        op_text += f'@Operation(summary = {safe_str(op_summary)}, description = {safe_str(op_description)})\n'
        funcname = make_funcname(operation["operationId"])
        # 打印@ApiResponses注解
        resp_ktype = 'String'
        multiple_resp = False
        resp_lines = []
        for status_code, resp_of_status_code in operation.get('responses', {}).items():
            content_lines = []
            for content_type, resp_of_content in resp_of_status_code.get('content', {}).items():
                type_prefix = funcname + ('Fail' if status_code >= '400' else 'Resp')
                votype = schema_to_votype(type_prefix, resp_of_content['schema'], swagger_data)
                content_lines.append(f'Content(mediaType = "{content_type}", schema = Schema(implementation = {votype}::class))')
                if status_code == '200' and 'application/json' in content_type.lower():
                    resp_ktype = schema_to_votype(type_prefix, resp_of_content['schema'], swagger_data, inherit=False)
                else:
                    multiple_resp = True
            resp_lines.append(f'ApiResponse(responseCode = "{status_code}", content = [{", ".join(content_lines)}])')
        if multiple_resp:
            op_text += '@ApiResponses(value = [\n'
            op_text += indented(',\n'.join(resp_lines)) + '\n])\n'
        # 打印函数定义和参数列表
        op_text += f'fun {funcname}('
        param_lines = []
        for source in ['path', 'query']:
            for item in parsed_op['params'][source]:
                name = item.get('name')
                if not name:
                    raise BadParamError(f'name不能为空', param=f'{path} {method.upper()} {source}')
                description = item.get('description', name)
                ktype = jsontype_to_ktype(item['schema']['type'], format=item['schema'].get('format', ''))
                if not item.get('required'):
                    ktype += '?'
                if source == 'path':
                    param_lines.append(f'@Parameter(description = {safe_str(description)}) @PathVariable {name}: {ktype}')
                elif source == 'query':
                    param_lines.append(f'@Parameter(description = {safe_str(description)}) {name}: {ktype}')
        if parsed_op['requests'] and 'application/json' in operation['requestBody']['content']:
            votype = schema_to_votype(funcname + 'Req', operation['requestBody']['content']['application/json']['schema'], swagger_data, inherit=False)
            param_lines.append(f'@RequestBody vo: {votype}')
        if len(param_lines) > 1:
            op_text += '\n' + indented(',\n'.join(param_lines))
        elif len(param_lines) == 1:
            op_text += param_lines[0]
        op_text += f'): {resp_ktype}?' + ' {\n    return null\n}'
        ret += indented(op_text) + '\n\n'
    return ret + '}\n'


def import_headers():
    return """package com.wax.controller
import io.swagger.v3.oas.annotations.Operation
import io.swagger.v3.oas.annotations.Parameter
import io.swagger.v3.oas.annotations.media.Content
import io.swagger.v3.oas.annotations.media.Schema
import io.swagger.v3.oas.annotations.responses.ApiResponse
import io.swagger.v3.oas.annotations.responses.ApiResponses
import java.time.LocalDate
import java.time.LocalDateTime
import java.time.LocalTime
import javax.validation.constraints.NotNull
import org.json.JSONObject
import org.springframework.validation.annotation.Validated
import org.springframework.web.bind.annotation.*\n\n"""


from functools import partial
from wax.lessweb.plugin.redisplugin import RedisServ
from wax.load_swagger import SwaggerData


def return_str(f):
    def g(*a, **b):
        ret = f(*a, **b)
        if isinstance(ret, bytes):
            return ret.decode()
        return ret
    return g


class PatialRedis:
    def __init__(self, name, redis):
        self._name = name
        self._redis = redis

    def __getattr__(self, method):
        return return_str(partial(getattr(self._redis, method), SwaggerData.redis_prefix + self._name))


class StateServ:
    redis_serv: RedisServ
    operation_id: str = ''

    @property
    def basic(self):  # format: 'statusCode:contentType:exampleName'
        return PatialRedis(name=f'{self.operation_id}::basic', redis=self.redis_serv.redis)

    @property
    def extra(self):
        return PatialRedis(name=f'{self.operation_id}::extra', redis=self.redis_serv.redis)



from typing import Optional, Dict, List, Union, TYPE_CHECKING
import json
import os

from requests.structures import CaseInsensitiveDict
from urllib.parse import unquote

from .webapi import Cookie, HttpStatus, ResponseStatus, ParamInput
from .bridge import Jsonizable, ParamStr, MultipartFile
from .webapi import header_name_of_wsgi_key, wsgi_key_of_header_name
from .webapi import parse_cookie, mimetypes
from .utils import eafp


__all__ = ["Request", "Response", "Context"]


if TYPE_CHECKING:
    from wax.lessweb.application import Application


class Request:
    """
    Contextual variables:
        * environ a.k.a. env – a dictionary containing the standard WSGI environment variables
        * home – the base path for the application, including any parts "consumed" by outer applications
        * homedomain – ? (appears to be protocol + host)
        * homepath – The part of the path requested by the user which was trimmed off the current app. That is homepath + path = the path actually requested in HTTP by the user. E.g. /admin This seems to be derived during startup from the environment variable REAL_SCRIPT_NAME. It affects what web.url() will prepend to supplied urls. This in turn affects where web.seeother() will go, which might interact badly with your url rewriting scheme (e.g. mod_rewrite)
        * host – the hostname (domain) and (if not default) the port requested by the user. E.g. example.org, example.org:8080
        * ip – the IP address of the user. E.g. xxx.xxx.xxx.xxx
        * method – the HTTP method used. E.g. POST
        * path – the path requested by the user, relative to the current application. If you are using subapplications, any part of the url matched by the outer application will be trimmed off. E.g. you have a main app in code.py, and a subapplication called admin.py. In code.py, you point /admin to admin.app. In admin.py, you point /stories to a class called stories. Within stories, web.ctx.path will be /stories, not /admin/stories.
        * protocol – the protocol used. E.g. https
        * query – an empty string if there are no query arguments otherwise a ? followed by the query string.
        * fullpath a.k.a. path + query – the path requested including query arguments but not including homepath.

        e.g. GET http://localhost:8080/api/hello/echo?a=1&b=2
            host => localhost:8080
            protocol => http
            homedomain => http://localhost:8080
            homepath => /api
            home => http://localhost:8080/api
            ip => 127.0.0.1
            method => GET
            path => /hello/echo
            query => a=1&b=2
            fullpath => /hello/echo?a=1&b=2

        lessweb use ctx.path in routing.
    """
    def __init__(self, encoding: str):
        self._cookies: Dict[str, str] = {}
        self._aliases: Dict[str, str] = {}  # alias {realname: queryname}
        self._params: Dict[str, Union[ParamStr, Jsonizable, None]] = {}

        self.encoding: str = encoding
        self.environ: Dict = {}
        self.env: Dict = {}
        self.host: str = ''
        self.protocol: str = ''
        self.homedomain: str = ''
        self.homepath: str = ''
        self.home: str = ''
        self.ip: str = ''
        self.method: str = ''
        self.path: str = ''  # ctx.path是路由的依据，因此是unquote之后的结果
        self.query: str = ''
        self.fullpath: str = ''

        self.body_data: Optional[bytes] = None  # Raw Body Input
        self.json_input: Optional[Jsonizable] = None  # Input from Json Body
        self.param_input: ParamInput = ParamInput()  # Param Inputs
        self.file_input: Dict[str, List[MultipartFile]] = {}  # Uploaded File Inputs

    def load(self, env):
        encoding = self.encoding
        request_uri = env.get('REQUEST_URI')
        self.environ = self.env = env
        self.host = env.get('HTTP_HOST', '[unknown]')
        if env.get('wsgi.url_scheme') in ['http', 'https']:
            self.protocol = env['wsgi.url_scheme']
        elif env.get('HTTPS', '').lower() in ['on', 'true', '1']:
            self.protocol = 'https'
        else:
            self.protocol = 'http'
        self.homedomain = self.protocol + '://' + self.host
        self.homepath = os.environ.get('REAL_SCRIPT_NAME', env.get('SCRIPT_NAME', ''))
        self.home = self.homedomain + self.homepath
        self.ip = env.get('REMOTE_ADDR', '0.0.0.0')
        self.method = env.get('REQUEST_METHOD', '')
        if request_uri is not None:
            self.path = unquote(request_uri.split('?', 1)[0][len(self.homepath):], encoding=encoding)
        else:
            self.path = env.get('PATH_INFO', '')  # you have to follow your server's default path encoding
        self.query = env.get('QUERY_STRING', '')
        self.fullpath = self.homedomain + (request_uri or '')
        # init cookie
        if not self._cookies and self.contains_header('cookie'):
            self._cookies = parse_cookie(self.get_header('cookie'))
        # parse query params
        self.param_input.load_query(self.query, encoding)
        # load body data
        cl = eafp(lambda: int(self.env.get('CONTENT_LENGTH')), 0)
        self.body_data = self.env['wsgi.input'].read(cl) if cl else None
        # parse form params
        if self.body_data:
            if self.is_json():
                self.json_input = eafp(lambda: json.loads(self.body_data.decode(encoding)),
                                       {'__error__': 'invalid json received'})
            elif self.is_form():
                eafp(lambda: self.param_input.load_form(self.body_data, self.env, encoding, self.file_input), None)


    def set_alias(self, realname, queryname):
        self._aliases[realname] = queryname

    def get_content_type(self) -> str:
        return self.env.get('CONTENT_TYPE', '')

    def is_json(self) -> bool:
        return 'json' in self.get_content_type().lower()

    def is_form(self) -> bool:
        content_type = self.get_content_type().lower()
        return bool(content_type) and ('form-' in content_type or '-urlencoded' in content_type)

    def contains_cookie(self, name: str) -> bool:
        return name in self._cookies

    def get_cookie(self, name: str) -> Optional[str]:
        return self._cookies.get(name)

    def get_cookienames(self) -> List[str]:
        return list(self._cookies.keys())

    def contains_header(self, name: str) -> bool:
        return wsgi_key_of_header_name(name) in self.env

    def get_header(self, name: str) -> Optional[str]:
        """
        根据http规范，多header应该合并入一个key/value，例如requests的headers就是dict。
        """
        return self.env.get(wsgi_key_of_header_name(name))

    def get_headernames(self) -> List[str]:
        return [s for s in (header_name_of_wsgi_key(k) for k in self.env.keys()) if s]

    def get_auth_bearer(self) -> Optional[str]:
        header_val = self.get_header('Authorization')
        if header_val is None:
            return None
        return header_val[len('Bearer '):]

    def get_input(self, key: str) -> Optional[Union[ParamStr, Jsonizable]]:
        # 注意cgi不保证重名参数的顺序，如果每次请求得到的结果可能不同
        if key in self._params:
            return self._params[key]
        param = self.param_input
        ret: Optional[Union[ParamStr, Jsonizable]]
        if key in param.url_input:
            self._params[key] = ret = param.url_input[key]
        elif key in param.form_input:  # must in front of query_input
            self._params[key] = ret = param.form_input[key][0]
        elif key in param.query_input:
            self._params[key] = ret = param.query_input[key][0]
        elif isinstance(self.json_input, dict):
            self._params[key] = ret = self.json_input.get(key, None)
        else:
            self._params[key] = ret = None
        return ret

    def get_uploaded_files(self, key: str) -> List[MultipartFile]:
        return self.file_input.get(key, [])


class Response:
    def __init__(self, encoding: str):
        self._cookies: Dict[str, Cookie] = {}
        self._status: Union[HttpStatus, ResponseStatus] = HttpStatus.OK
        self._headers: CaseInsensitiveDict = CaseInsensitiveDict()
        self.encoding: str = encoding

    def set_cookie(self, name:str, value:str, expires:int=None, path:str='/',
                   domain:str=None, secure:bool=False, httponly:bool=False) -> None:
        self._cookies[name] = Cookie(name, value, expires, path, domain, secure, httponly)

    def get_cookie(self, name:str) -> Optional[Cookie]:
        return self._cookies.get(name)

    def del_cookie(self, name:str) -> None:
        self._cookies.pop(name, None)

    def set_status(self, status: Union[HttpStatus, ResponseStatus]) -> None:
        self._status = status

    def get_status(self) -> Union[HttpStatus, ResponseStatus]:
        return self._status

    def set_header(self, name: str, value: Union[str, int]) -> None:
        if isinstance(value, int):
            self._headers[name] = str(value)
        elif '\n' in name or '\r' in name or '\n' in value or '\r' in value:
            raise ValueError('invalid characters in header')
        else:
            self._headers[name] = value

    def get_header(self, name: str) -> Optional[str]:
        return self._headers.get(name)

    def del_header(self, name: str) -> str:
        return self._headers.pop(name, None)

    def get_headernames(self) -> List[str]:
        return list(self._headers.keys())

    def clear(self) -> None:
        """
        Clear headers and cookies.
        """
        self._headers.clear()
        self._cookies.clear()

    def send_access_allow(self, allow_headers: List[str]=None) -> None:
        allow_headers = allow_headers or []
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, PUT, DELETE, OPTIONS')
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers',
                        'Cache-Control, Accept-Encoding, Origin, X-Requested-With, Content-Type, Accept, '
                        'Authorization, Referer, User-Agent' + ''.join(', '+h for h in allow_headers))

    def send_allow_methods(self, methods: List[str]):
        self.set_header('Allow', ', '.join(methods))

    def send_redirect(self, location: str) -> None:
        self.set_header('Location', location)

    def send_content_type(self, mimekey='html', encoding: str=''):
        mimekey = mimekey.lower()
        if encoding:
            self.set_header('Content-Type', '%s; charset=%s' % (mimetypes[mimekey], encoding))
        else:
            self.set_header('Content-Type', '%s' % mimetypes[mimekey])


class Context(object):
    def __init__(self, app: 'Application') -> None:
        self.app_stack: List = []
        self.app: Application = app
        self.request: Request = Request(app.encoding)
        self.response: Response = Response(app.encoding)
        self.box: Dict = {}

    def __call__(self):
        return self.app_stack[-1](self)

